@@ -10,6 +10,7 @@ import time
 import urllib.error
 import urllib.parse
 import urllib.request
+import zipfile
 from copy import deepcopy
 from http.client import IncompleteRead
 from inspect import isfunction
@@ -51,6 +52,9 @@ class MixpanelUtils(object):
         self,
         api_secret,
         token=None,
+        service_account_username=None,
+        project_id=None,
+        strict_import=True,
         timeout=120,
         pool_size=None,
         read_pool_size=2,
@@ -60,8 +64,12 @@ class MixpanelUtils(object):
     ):
         """Initializes the MixpanelUtils object
 
-        :param api_secret: API Secret for your project
+        :param api_secret: API Secret for your project OR your Service Account
         :param token: Project Token for your project, required for imports
+        :param service_account_username: Username for your Service Account
+        :param project_id: project id, required for Service Account authentication
+        :param strict_import: When set to True (recommended), Mixpanel will validate imported events and return errors
+            per event that failed. (Default value = True)
         :param timeout: Time in seconds to wait for HTTP responses
         :param pool_size: Number of threads to use for sending data to Mixpanel (Default value = cpu_count * 2)
         :param read_pool_size: Separate number of threads to use just for read operations (i.e. query_engage)
@@ -71,6 +79,9 @@ class MixpanelUtils(object):
         :param eu: Is the project participating in EU residency
         :type api_secret: str
         :type token: str
+        :type service_account_username: str
+        :type project_id: int
+        :type strict_import: bool
         :type timeout: int
         :type pool_size: int
         :type read_pool_size: int
@@ -82,6 +93,11 @@ class MixpanelUtils(object):
 
         self.api_secret = api_secret
         self.token = token
+        self.service_account_username = service_account_username
+        self.project_id = project_id
+        self.strict_import = strict_import
+        if self.service_account_username is not None:
+            assert self.project_id, "project_id required for Service Account authentication!"
         self.timeout = timeout
         if pool_size is None:
             # Default number of threads is system dependent
@@ -203,36 +219,43 @@ class MixpanelUtils(object):
                 base = [base_url]
             else:
                 base = [base_url, str(MixpanelUtils.VERSION)]
-
             request_url = "/".join(base + path_components)
 
-            encoded_params = MixpanelUtils._unicode_urlencode(params)
+            if self.service_account_username:
+                basic_credentials = f"{self.service_account_username}:{self.api_secret}"
+            else:
+                basic_credentials = f"{self.api_secret}:"
+            encoded_credentials = base64.b64encode(basic_credentials.encode("utf-8")).decode("utf-8")
+
+            if headers is None:
+                headers = {}
+            headers["Authorization"] = f"Basic {encoded_credentials}"
 
             # Set up request url and body based on HTTP method and endpoint
             if method == "GET" or method == "DELETE":
                 data = None
-                request_url += "?" + encoded_params
+                if self.service_account_username:
+                    params['project_id'] = self.project_id
+                request_url += "?" + MixpanelUtils._unicode_urlencode(params)
             else:
-                data = encoded_params
-                if (
-                    base_url == self.import_api
-                    or "import-people" in path_components
-                    or "import-events" in path_components
-                ):
-                    data += "&verbose=1"
+                if "engage" in path_components:
+                    data = MixpanelUtils._unicode_urlencode(params).encode("utf-8")
+                    request_url += "?verbose=1"
+                else:
+                    headers["Content-Type"] = "application/json"
+                    data = params["data"]
+                    query_params = {}
+                    if self.strict_import:
+                        query_params["strict"] = 1
+                    if self.service_account_username:
+                        query_params["project_id"] = self.project_id
+                    if query_params:
+                        request_url += "?" + MixpanelUtils._unicode_urlencode(query_params)
                 # Uncomment the line below to debug log the request body data
                 # MixpanelUtils.LOGGER.debug(f"{method} data: {data}")
-                data = data.encode("utf-8")
             MixpanelUtils.LOGGER.debug(f"Request Method: {method}")
             MixpanelUtils.LOGGER.debug(f"Request URL: {request_url}")
 
-            if headers is None:
-                headers = {}
-            headers["Authorization"] = "Basic {encoded_secret}".format(
-                encoded_secret=base64.b64encode(
-                    "{}:".format(self.api_secret).encode("utf-8")
-                ).decode("utf-8")
-            )
             request = urllib.request.Request(request_url, data, headers, method=method)
             MixpanelUtils.LOGGER.debug(f"Request Headers: {json.dumps(headers)}")
 
@@ -241,11 +264,11 @@ class MixpanelUtils(object):
                 if raw_stream and base_url == self.raw_api:
                     return response
             except urllib.error.HTTPError as e:
-                MixpanelUtils.LOGGER.warning("The server couldn't fulfill the request.")
-                MixpanelUtils.LOGGER.warning(f"Error code: {e.code}")
-                MixpanelUtils.LOGGER.warning(f"Reason: {e.reason}")
+                MixpanelUtils.LOGGER.error("The server couldn't fulfill the request.")
+                MixpanelUtils.LOGGER.error(f"HTTP Error Code: {e.code}")
+                MixpanelUtils.LOGGER.error(f"Reason: {e.reason}")
                 if hasattr(e, "read"):
-                    MixpanelUtils.LOGGER.warning(f"Response: {e.read()}")
+                    MixpanelUtils.LOGGER.error(f"Response: {e.read().decode('utf-8')}")
                 if e.code >= 500:
                     # Retry if we get an HTTP 5xx error
                     MixpanelUtils.LOGGER.warning(f"Attempting retry #{retries + 1}")
@@ -258,14 +281,12 @@ class MixpanelUtils(object):
                         raw_stream=raw_stream,
                         retries=retries + 1,
                     )
-                else:
-                    raise
 
             except urllib.error.URLError as e:
-                MixpanelUtils.LOGGER.warning("We failed to reach a server.")
-                MixpanelUtils.LOGGER.warning(f"Reason: {e.reason}")
+                MixpanelUtils.LOGGER.error("We failed to reach a server.")
+                MixpanelUtils.LOGGER.error(f"Reason: {e.reason}")
                 if hasattr(e, "read"):
-                    MixpanelUtils.LOGGER.warning(f"Response: {e.read()}")
+                    MixpanelUtils.LOGGER.error(f"Response: {e.read()}")
                 MixpanelUtils.LOGGER.warning(f"Attempting retry #{retries + 1}")
                 return self.request(
                     base_url,
@@ -277,7 +298,7 @@ class MixpanelUtils(object):
                     retries=retries + 1,
                 )
             except timeout:
-                MixpanelUtils.LOGGER.warning("The read operation timed out.")
+                MixpanelUtils.LOGGER.error("The read operation timed out.")
                 self.timeout = self.timeout + 30
                 MixpanelUtils.LOGGER.warning(
                     f"Increasing timeout to {self.timeout} and attempting retry #{retries + 1}"
@@ -300,7 +321,7 @@ class MixpanelUtils(object):
                         response_data = response.read()
                     return response_data.decode("utf-8")
                 except IncompleteRead:
-                    MixpanelUtils.LOGGER.warning(
+                    MixpanelUtils.LOGGER.error(
                         f"Response data is incomplete. Attempting retry #{retries + 1}"
                     )
                     return self.request(
@@ -313,7 +334,7 @@ class MixpanelUtils(object):
                         retries=retries + 1,
                     )
         else:
-            MixpanelUtils.LOGGER.warning(
+            MixpanelUtils.LOGGER.error(
                 "Maximum retries reached. Request failed. Try again later."
             )
             raise BaseException
@@ -357,7 +378,7 @@ class MixpanelUtils(object):
         """
         assert self.token, "Project token required for People operation!"
         if profiles is not None and query_params is not None:
-            MixpanelUtils.LOGGER.warning(
+            MixpanelUtils.LOGGER.error(
                 "profiles and query_params both provided, please use one or the other"
             )
             return
@@ -1418,9 +1439,7 @@ class MixpanelUtils(object):
         :type timezone_offset: int | float
 
         """
-        self._import_data(
-            data, self.import_api, "import", timezone_offset=timezone_offset, batch_size=2000
-        )
+        self._import_data(data, self.import_api, "import", timezone_offset=timezone_offset)
 
     def import_people(self, data, ignore_alias=False, raw_record_import=False):
         """Imports a list of Mixpanel People profile dicts (or raw API update operations) or a file containing a JSON
@@ -1444,6 +1463,27 @@ class MixpanelUtils(object):
             raw_record_import=raw_record_import,
         )
 
+    def import_from_amplitude(self, amplitude_api_key, amplitude_api_secret, start, end):
+        """Exports all data from an Amplitude project and imports it into Mixpanel
+
+            :param amplitude_api_key: Your Amplitude API key
+            :param amplitude_api_secret: Your Amplitude API secret
+            :param start: Date and time in the format of YYYYMMDDTHH (e.g. '20150201T05')
+            :param end: Date and time in the format of YYYYMMDDTHH (e.g. '20150203T20')
+
+            :type amplitude_api_key: string
+            :type amplitude_api_secret: string
+            :type start: string
+            :type end: string
+        """
+        amplitude_export_url = f"https://amplitude.com/api/2/export?start={start}&end={end}"
+        credentials = f"{amplitude_api_key}:{amplitude_api_secret}"
+
+        extract_data_path = self._extract_amplitude_data(amplitude_export_url, credentials)
+        self._transform_and_load_amplitude_data(extract_data_path)
+
+        return
+
     """
     Private, internal methods
     """
@@ -1477,19 +1517,26 @@ class MixpanelUtils(object):
         """Takes a Mixpanel API response and checks the status
 
         Logs a warning message if status is not equal to 1.
-        *Does not* raise an exception on error: if a callback in a ThreadPool apply_async
-        call throws an exception the pool will hang indefinitely
 
         :param response: A Mixpanel API JSON response
         :type response: str
 
         """
-        response_data = json.loads(response)
-        if ("status" in response_data and response_data["status"] != 1) or (
-            "status" not in response_data
-        ):
-            MixpanelUtils.LOGGER.warning(f"Bad API response: {response}")
         MixpanelUtils.LOGGER.debug(f"API Response: {response}")
+        if response is not None:
+            try:
+                response_data = json.loads(response)
+                if "status" in response_data:
+                    if response_data["status"] != 1 and response_data["status"] != "OK":
+                        MixpanelUtils.LOGGER.warning(f"API response NOT OK: {response}")
+                else:
+                    MixpanelUtils.LOGGER.warning(f"API response NO STATUS: {response}")
+            except TypeError:
+                if response != "1":
+                    MixpanelUtils.LOGGER.warning(f"API response NOT OK: {response}")
+            except BaseException:
+                MixpanelUtils.LOGGER.error("Exception in _async_response_handler_callback!", exc_info=True)
+                raise
 
     @staticmethod
     def _write_items_to_csv(items, output_file):
@@ -1510,7 +1557,7 @@ class MixpanelUtils(object):
                 props_key = "properties"
                 initial_header_value = "event"
         else:
-            MixpanelUtils.LOGGER.warning("No data to write!")
+            MixpanelUtils.LOGGER.error("No data to write!")
             return
 
         columns = [list(item[props_key].keys()) for item in items]
@@ -1665,7 +1712,7 @@ class MixpanelUtils(object):
         elif isinstance(arg, list):
             item_list = arg
         else:
-            MixpanelUtils.LOGGER.warning(
+            MixpanelUtils.LOGGER.error(
                 "data parameter must be a string filename or a list of items"
             )
 
@@ -1711,7 +1758,7 @@ class MixpanelUtils(object):
                             )
                             item_list.append(profile)
                     else:
-                        MixpanelUtils.LOGGER.warning(
+                        MixpanelUtils.LOGGER.error(
                             "Unable to determine Mixpanel data type: CSV header does not contain 'event' or '$distinct_id'"
                         )
                         return None
@@ -1881,7 +1928,7 @@ class MixpanelUtils(object):
             if compress:
                 MixpanelUtils._gzip_file(output_file)
         else:
-            MixpanelUtils.LOGGER.warning(
+            MixpanelUtils.LOGGER.error(
                 f"Invalid format must be either json or csv, got: {format}"
             )
             return
@@ -1900,11 +1947,11 @@ class MixpanelUtils(object):
         if "results" in data:
             return data
         else:
-            MixpanelUtils.LOGGER.warning(f"Invalid response from /engage: {response}")
+            MixpanelUtils.LOGGER.error(f"Invalid response from /engage: {response}")
             return
 
     def _dispatch_batches(
-        self, base_url, endpoint, item_list, prep_args, batch_size=50
+        self, base_url, endpoint, item_list, prep_args, batch_size=2000
     ):
         """Asynchronously sends batches of items to the /import, /engage, /import-events or /import-people Mixpanel API
         endpoints
@@ -1929,7 +1976,7 @@ class MixpanelUtils(object):
         elif endpoint == "engage" or endpoint == "import-people":
             prep_function = MixpanelUtils._prep_params_for_profile
         else:
-            MixpanelUtils.LOGGER.warning(
+            MixpanelUtils.LOGGER.error(
                 f'endpoint must be "import", "engage", "import-events" or "import-people", found: {endpoint}'
             )
             return
@@ -1982,7 +2029,10 @@ class MixpanelUtils(object):
 
         """
         try:
-            params = {"data": base64.b64encode(json.dumps(batch).encode("utf-8"))}
+            data = json.dumps(batch).encode("utf-8")
+            if endpoint != "import":
+                data = base64.b64encode(data)
+            params = {"data": data}
             response = self.request(
                 base_url, [endpoint], params, "POST", retries=retries
             )
@@ -2007,7 +2057,7 @@ class MixpanelUtils(object):
         timezone_offset=None,
         ignore_alias=False,
         raw_record_import=False,
-        batch_size=50
+        batch_size=2000
     ):
         """Base method to import either event data or People profile data as a list of dicts or from a JSON array
         file
@@ -2105,7 +2155,7 @@ class MixpanelUtils(object):
             elif isinstance(event_selectors, list):
                 pass
             else:
-                MixpanelUtils.LOGGER.warning(
+                MixpanelUtils.LOGGER.error(
                     f"Invalid type for event_selectors, must be dict or list, found: {type(event_selectors)}"
                 )
 
@@ -2131,14 +2181,14 @@ class MixpanelUtils(object):
             elif isinstance(user_selectors, list):
                 pass
             else:
-                MixpanelUtils.LOGGER.warning(
+                MixpanelUtils.LOGGER.error(
                     f"Invalid type for user_selectors, must be str or list, found: {type(user_selectors)}"
                 )
                 return
 
             params = {"user_selectors": user_selectors}
         else:
-            MixpanelUtils.LOGGER.warning(
+            MixpanelUtils.LOGGER.error(
                 f'Invalid data_type, must be "events" or "people", found: {data_type}'
             )
             return
@@ -2147,3 +2197,172 @@ class MixpanelUtils(object):
             params["output_properties"] = output_properties
 
         return self.query_jql(jql_script, params=params, format=format)
+
+    def _create_merge_event(self, event):
+        return {
+            "event": "$merge",
+            "properties": {
+                "$distinct_ids": [
+                    event["user_id"],
+                    str(event["amplitude_id"])
+                ],
+                "time": int(time.time() * 1000),
+                "distinct_id": ""
+            }
+        }
+
+    def _map_amplitude_property_to_mixpanel(self, property_name):
+        amplitude_to_mixpanel_map = {
+            "app_version": "$app_version_string",
+            "os_name": "$os",
+            "os_version": "$os_version",
+            "device_brand": "$brand",
+            "device_manufacturer": "$manufacturer",
+            "device_model": "$model",
+            "carrier": "$carrier",
+            "region": "$region",
+            "city": "$city",
+            "insert_id": "$insert_id",
+            "platform": "platform",
+            "dma": "dma",
+            "language": "language",
+            "price": "price",
+            "quantity": "quantity",
+            "revenue": "revenue",
+            "productId": "productId",
+            "revenueType": "revenueType",
+            "location_lat": "location_lat",
+            "location_lng": "location_lng",
+            "idfa": "idfa",
+            "idfv": "idfv",
+            "adid": "adid",
+            "android_id": "android_id",
+            "event_id": "event_id",
+            "session_id": "session_id",
+            "plan": "plan"
+        }
+        return amplitude_to_mixpanel_map.get(property_name)
+
+    def _transform_amplitude_profiles(self, amplitude_profile):
+        properties = amplitude_profile["user_properties"]
+        default_properties = {self._map_amplitude_property_to_mixpanel(key):value for key, value in amplitude_profile.items() if self._map_amplitude_property_to_mixpanel(key)}
+        profile = {
+            "$token": self.token,
+            "$distinct_id": amplitude_profile["user_id"],
+            "$ip": amplitude_profile["ip_address"],
+            "$properties": {**properties, **default_properties}
+        }
+
+        return profile
+
+
+    def _format_amplitude_time(self, event_time):
+        for date_format in ('%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S.%f'):
+            try:
+                return datetime.datetime.strptime(event_time, date_format)
+            except ValueError:
+                pass
+        raise ValueError('No valid date format found')
+
+
+    def _transform_amplitude_events(self, amplitude_event):
+        event_dt = self._format_amplitude_time(amplitude_event["event_time"])
+
+        mixpanel_properties = {
+            # prefer user_id, then device_id, then amplitude_id
+            "distinct_id": amplitude_event.get("user_id") or amplitude_event.get("device_id") or amplitude_event["amplitude_id"],
+            "$device_id": amplitude_event["device_id"],
+            "time": int(event_dt.timestamp() * 1000),
+            "ip": amplitude_event["ip_address"],
+            "mp_country_code": amplitude_event["country"]
+        }
+
+        default_properties = {self._map_amplitude_property_to_mixpanel(key):value for key, value in amplitude_event.items() if self._map_amplitude_property_to_mixpanel(key)}
+
+        combined_properties = {**mixpanel_properties, **amplitude_event["event_properties"], **default_properties}
+
+        event = {
+            "event": amplitude_event["event_type"],
+            "properties": combined_properties
+        }
+
+        return event
+
+    def _dedupe_merge_events(self, merge_events):
+        unique_merge_events = {}
+        for event in merge_events:
+            merge_pair = (event["properties"]["$distinct_ids"][0], event["properties"]["$distinct_ids"][1])
+            if not unique_merge_events.get(merge_pair):
+                unique_merge_events[merge_pair] = event
+
+        return list(unique_merge_events.values())
+
+    def _extract_amplitude_data(self, url, credentials):
+        zip_file_path = "./amp_data.zip"
+        data_parent_path = "./amp_data"
+        extract_path = f"{data_parent_path}/amplitude_extract"
+        req = urllib.request.Request(url)
+        encoded_credentials = base64.b64encode(credentials.encode('ascii'))
+        req.add_header('Authorization', 'Basic %s' % encoded_credentials.decode("ascii"))
+
+        try:
+            # Download zip file
+            response = urllib.request.urlopen(req)
+            with open(zip_file_path, "wb") as zip_file:
+                zip_file.write(response.read())
+
+            # Unzip file
+            with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
+                zip_ref.extractall(data_parent_path)
+
+            data_path = next(os.walk(data_parent_path))[1][0]
+            full_data_path = os.path.join(data_parent_path, data_path)
+
+            extracted_event_count = 0
+
+            # Parse all .gz data files into json files
+            for filename in os.listdir(full_data_path):
+                if filename[-3:] == ".gz":
+                    with gzip.open(os.path.join(full_data_path, filename), 'rb') as f:
+                        events = []
+                        for line in f:
+                            events.append(json.loads(line))
+                        extracted_event_count += len(events)
+                        os.makedirs(extract_path, exist_ok=True)
+                        with open(os.path.join(extract_path, filename[:-3]), "w") as extract_file:
+                            extract_file.write(json.dumps(events))
+
+            return extract_path
+        except Exception as e:
+            MixpanelUtils.LOGGER.error(
+                "Error extracting data from Amplitude", exc_info=True
+            )
+            raise e
+
+    def _transform_and_load_amplitude_data(self, extract_data_path):
+        transform_data_path = "./amp_data/amplitude_transform"
+        try:
+            total_events = 0
+            for filename in os.listdir(extract_data_path):
+                all_events = []
+                with open(os.path.join(extract_data_path, filename), "r") as extract_file:
+                    all_events = json.loads(extract_file.read())
+
+                transformed_profiles = [self._transform_amplitude_profiles(profile) for profile in all_events if profile["user_properties"]]
+                transformed_events = [self._transform_amplitude_events(event) for event in all_events]
+                merge_events = [self._create_merge_event(event) for event in all_events if event.get("user_id") and event.get("amplitude_id")]
+
+                unique_merge_events = self._dedupe_merge_events(merge_events)
+
+                self.import_people(transformed_profiles)
+                self.import_events(transformed_events, 0)
+                self.import_events(unique_merge_events, 0)
+                total_events += len(all_events)
+
+            print(f"Imported {total_events} events")
+
+        except:
+            MixpanelUtils.LOGGER.error(
+                "Error transforming Amplitude data", exc_info=True
+            )
+            return
